#include "EnergyModel.h"
ostream & operator << (ostream& out, Parameter & p){
	out<<"> E:"<<endl;
	for(int i=0; i< p.l(); i++){
		for(int j=0; j<4; j++){ out<<p.E(i,j)<<" "; }
		out<<endl;
	}
	out<<"> D:"<<endl;
	for(int i=0; i< p.l(); i++){
		for(int j=0; j<4; j++){ out<<p.D(i,j)<<" "; }
		out<<endl;
	}
	out<<"> Eopt= "<<p.Eopt()<<endl;
	out<<"> lntf= "<<p.lntf()<<endl;
	
	out<<"> S:"<<endl;
	for(int i=0; i< p.n(); i++){
		for(int j=0; j<p.l(); j++) out<<p.S(i,j);
		out<<endl;
	}
	out<<"> R:"<<endl;
	for(int i=0; i< p.n(); i++){ out<<p.R(i)<<" "; } out<<endl;
	out<<"> C:"<<endl;
	for(int i=0; i< p.n(); i++){ out<<p.C(i)<<" "; } out<<endl;
	out<<"> B:"<<endl;
	for(int i=0; i< p.n(); i++){ out<<p.B(i)<<" "; } out<<endl;
	out<<"> Btot= "<<p.Btot()<<endl;
	out<<"> Ctot= "<<p.Ctot()<<endl;
	out<<"> W:"<<endl;
	for(int i=0; i< p.n(); i++){ out<<p.W(i)<<" "; } out<<endl;
	out<<"> F:"<<endl;
	for(int i=0; i< p.n(); i++){ out<<p.F(i)<<" "; } out<<endl;
	
	out<<"> GE:"<<endl;
	for(int i=0; i< p.n(); i++){ out<<p.GE(i)<<" "; } out<<endl;
	out<<"> G:"<<endl;
	for(int i=0; i< p.n(); i++){ out<<p.G(i)<<" "; } out<<endl;
	out<<"> P:"<<endl;
	for(int i=0; i< p.n(); i++){ out<<p.P(i)<<" "; } out<<endl;
}

ostream & operator << (ostream& out, EModel & em){
}

Parameter::Parameter(int n, int l){
	Init(n,l);
}
Parameter::Parameter(Parameter& c){
	Init(c.n(),c.l());
	Copy(c);
}
Parameter::Parameter(string & fname){
	if(CheckInput(fname)){
		Init(n(),l());
		Read(fname);
	}else{
		cerr<<"ERROR: currupted file : "<<fname<<endl;
		Destroy();
		exit(-1);
	}
}

void Parameter::Read(string & fname){
	ifstream ifs(fname.c_str());
	
	vector<int> refids;
	vector<double> kds;
	string buf = "";
	//bool first = true;
	int n=0;
	for(int i=0; i< l()*4; i++) _D[i] = -1;// not use flag
	while(getline(ifs,buf)){
		if(buf[0] == '#') continue;
		if(buf.substr(0,7) == "> REFID"){
			// HK had declared 'kd' as int!  This is the likely problem.
			int refid;
			double kd;
			while(getline(ifs,buf)){
				if(buf=="//") break;
				istringstream iss(buf);
				iss>>refid>>kd;
				refids.push_back(refid);
				kds.push_back(kd);
			}
		}else if(buf.substr(0,5) == "> SEQ"){
			int C;
			int B;	
			string seq;
			while(getline(ifs,buf)){
				if(buf=="//") break;
				istringstream iss(buf);
				iss>>seq>>C>>B;	
				for(int i=0; i<l(); i++){
					int char2int = 0;
					if(seq[i]=='a' || seq[i]=='A'){
						char2int = 0;
					}else if(seq[i]=='c' || seq[i]=='C'){
						char2int = 1;
					}else if(seq[i]=='g' || seq[i]=='G'){
						char2int = 2;
					}else if(seq[i]=='t' || seq[i]=='T'){
						char2int = 3;
					}else{
						cerr<<"ERROR: unknown char: "<<seq[i]<<endl;
					}
					SetS(n,i,char2int);
					SetD(i,char2int,0);
				}
				_C[n] = C; _Ctot += C;
				_B[n] = B; _Btot += B;
				_G[n] = 0;
				_P[n] = 0;
				_R[n] = false;
				for(int m=0; m<refids.size(); m++){ 
					if(n == refids[m]){
						//	_G[n] = _RT * logarithm(kds[m]);
						_G[n] = logarithm(kds[m]);
						_R[n] = true;
						cout << "Setting oligo ";
						for(int i=0; i<l(); i++){
						    cout << seq[i];
						}
						cout << " as REF OLIGO to " << _G[n] << ", index " << n << endl;
					}
				}
				n++;
			}
		}
	}
	for(int i=0; i< l() ; i++) 
		for(int j=0; j< 4 ; j++) if(CntS(i,j) == n ) SetD(i,j,-1);
	
	double minVal = 0.0000000000001;
	// Setting the frequency prior to binding, based on prior counts and prior totals (W[i])
	for(int i=0; i< n ; i++) {
		_W[i] = _C[i]/(double)_Ctot;
		if (_W[i] <= minVal ) {
			_W[i] = minVal;
		}
	}
	
	for(int i=0; i< n ; i++) {
		_F[i] = _B[i]/(double)_Btot;
		if (_F[i] <= minVal ) {
			_F[i] = minVal;
		}
	}	
	ifs.close();
}
bool Parameter::CheckInput(string & fname){
	ifstream ifs(fname.c_str());
	string buf = "";
	double C;
	int B;	
	string seq;
	bool first = true;
	int n = 0;
	int l = 0;
	while(getline(ifs,buf)){
		if(buf[0] == '#') continue;
		if(buf.substr(0,5) == "> SEQ"){ 
			while(getline(ifs,buf)){
				if(buf == "//") break;
				istringstream iss(buf);
				iss>>seq>>C>>B;	
				if(first){
					l = seq.size(); n = 1;
					first = false;
				}else{
					if(l != seq.size()){
						cerr<<"ERROR: not equal seq. length in "<<fname<<endl;
						ifs.close();
						return(false);
					}
					n++;
				}
			}}
	}
	ifs.close();
	_n = n;
	_l = l;
	return(true);
}

void Parameter::Init(int n, int l){
	// R = 8.31447215 J/(K*mol) = 1.987/1000 kCal/(K*mol)
	_RT = 288*1.987/1000; // Kcal/(K*mol)
	
	_n = n;
	_l = l;
	if(_n == 0 || _l == 0){
		_E=NULL;_D=NULL;
		_S=NULL;_R=NULL;
		_C=NULL;_B=NULL;_P=NULL;_F=NULL;_G=NULL;_GE=NULL;
	}else{
		_E = new double [_l*4];
		_D = new int[_l*4];
		
		_S = new int[_n*_l];
		_R = new bool[_n];
		_W = new double[_n];
		_B = new int[_n];
		_C = new int[_n];
		_P = new double[_n]; 
		_F = new double[_n];
		_G = new double[_n];
		_GE = new double[_n];
		
		_lntf = 0;
		for(int i=0; i<_l*4; i++){
			_E[i] = 0.0;
			_D[i] = -1;//not using
		}
		for(int i=0; i<_n; i++){
			_C[i] = 0;
			_B[i] = 0;
			_W[i] = 0; 
			_R[i] = false;
			_G[i] = 0.0;
			_GE[i] = 0.0;
			_F[i] = 0.0;
			_P[i] = 0.0;
		}
		for(int i=0; i<_n*_l; i++){
			_S[i] = 0;
		}
	}
	_Eopt = 0;
	_Ctot = 0;
	_Btot = 0;
	_sigE = 0;
}
bool Parameter::Birth(int i, int j, double jp){
	cout << "Birth" << endl;
	int dij = D(i,j);
	if( dij == -1) return(false);
	bool single = (CntD(i,dij) == 1);
	if( dij == 0 && single) return(false);
	double left = E(i, FndD(i,dij-1));
	int iRight = FndD(i,dij+2);
	if(iRight > -1) jp = E(i,iRight)-left;
	
	SetD(i,j,++dij);
	for(int k=0; k<4; k++){
		int dik = D(i,k);
		if(dik==dij) SetE(i,k, left + _rng.uniform(0,jp));
		if(single && dik >= dij) SetD(i,k,--dik);
	}
	
	for(int m=0; m<n(); m++){
		if(D(i, S(m,i)) == dij) UpdateGE(m);
	}
	Normalize();
	return(true);
}

bool Parameter::Death(int i, int j, double jp){
    cout << "Death" << endl;
    
    int dij = D(i,j);
	if( dij < 1) return(false);
	int iLeft = FndD(i, dij-1);	
	if(E(i,iLeft)+jp < E(i,j)) return(false);
	bool single = (CntD(i,dij) == 1);
	SetE(i,j, E(i,iLeft));
	SetD(i,j,--dij);
	if(single){
		for(int k=0;k<4; k++)
		 	if(D(i,k) > dij) SetD(i,k, D(i,k)-1); 
	}
	for(int m=0; m<n(); m++){
		if(S(m,i) == j) UpdateGE(m);
	}
	Normalize();
	return(true);
}
double Parameter::Constrain(int i, int j, double v){
	int dij = D(i,j);
	int left = FndD(i, dij-1);
	double vLeft = E(i,left);
	if(left > -1 && v < vLeft){
		v = vLeft; 
	}
	return(v);
}

double Parameter::CalcDeltaE(int n){
	double e = 0;
	for(int j=0; j< _l; j++) { e += E(j,S(n,j)); }
	return(e);
}
void Parameter::Copy(Parameter & c){
	if(_n != c.n() || _l != c.l()){
		Destroy();
		Init(c.n(), c.l());
	}
	for(int i=0; i<_l; i++){
		for(int j=0; j<4; j++){
			SetE(i,j,c.E(i,j));	
			SetD(i,j,c.D(i,j));	
		}}
	
	SetBtot( c.Btot());
	SetCtot( c.Ctot());
	SetPave( c.Pave());
	SetEopt( c.Eopt());
	SetSigE( c.sigE());
	SetLntf( c.lntf());
	
	for(int i=0; i<_n; i++){
		// [blocking this - jdk]	for(int j=0; j<_l; j++) SetS(i,j, c.S(i,j));
		SetR(i, c.R(i));
		SetB(i, c.B(i));
		SetC(i, c.C(i));
		SetW(i, c.W(i));
		SetP(i, c.P(i));
		SetF(i, c.F(i));
		SetG(i, c.G(i));
		SetGE(i, c.GE(i));
	}
}
void Parameter::Destroy(){
	if(_E != NULL) delete _E; _E = NULL;
	if(_D != NULL) delete _D; _D = NULL;
	
	if(_W != NULL) delete _W; _W = NULL;
	if(_B != NULL) delete _B; _B = NULL;
	if(_C != NULL) delete _C; _C = NULL;
	if(_S != NULL) delete _S; _S = NULL;
	if(_R != NULL) delete _R; _R = NULL;
	if(_P != NULL) delete _P; _P = NULL;
	if(_F != NULL) delete _F; _F = NULL;
	if(_G != NULL) delete _G; _G = NULL;
	if(_GE != NULL) delete _GE; _GE = NULL;
}
Parameter::~Parameter(){
	Destroy();
}
EModel::EModel(Input * inp, Parameter * par){
	_ctrl = inp;
	_param = par;
	_fixdmat = false;
	
	printf("Initializing log lookup table...\n");
	initLookup(10000000);
	printf("Test: log(1.4) = %f, LUT log(1.4) = %f\n", log(1.4), logarithm(1.4) );
	
	if(_ctrl->_dmat.size() > 0){
		for(int i=0; i<_param->l(); i++){
			for(int j=0; j<4 ;j++){
				if(_param->D(i,j) >= 0){
					_param->SetD(i,j, _ctrl->_dmat[i*4+j]);
				}
			}}
		_fixdmat = true;
		for(int i=0; i<_param->n(); i++) _param->UpdateGE(i);
		_param->Normalize();
		cout<<"use fixed DMAT"<<endl;
	}
	
	_pwG = new PropWind(_ctrl->_g_win, _ctrl->_g_min, _ctrl->_g_max);
	_pwW = new PropWind(_ctrl->_w_win, _ctrl->_w_min, _ctrl->_w_max);
	_pwE = new PropWind(_ctrl->_e_win, _ctrl->_e_min, _ctrl->_e_max);
	_pwEopt = new PropWind(_ctrl->_eopt_win, _ctrl->_eopt_min, _ctrl->_eopt_max);
	_pwLntf = new PropWind(_ctrl->_lntf_win, _ctrl->_lntf_min, _ctrl->_lntf_max);
	_pwSigE = new PropWind(_ctrl->_s_win, _ctrl->_s_min, _ctrl->_s_max);
	
	_param->SetSigE(_ctrl->_s_ini);
	_param->SetEopt(_ctrl->_eopt_ini);
	
	_param->SetLntf(_ctrl->_lntf_ini);
	for(int m=0; m<_param->n(); m++){
		if(!_param->R(m))
			_param->SetG(m, _ctrl->_g_ini);
		_param->UpdateP(m);
	}
	_param->Normalize();
	_param_c = new Parameter(*_param);
	_param_c2 = new Parameter(*_param);
}

void EModel::Destroy(){
	delete _param_c;
	delete _param_c2;
	delete _pwG;
	delete _pwW;
	delete _pwE;
	delete _pwEopt;
	delete _pwLntf;
	delete _pwSigE;
}

EModel::~EModel(){
	Destroy();
}
void EModel::Read(string & fname){
	ifstream ifs(fname.c_str());
	string buf = "";
	bool first = true;
	vector<bool> marker;
	while(getline(ifs,buf)){
		istringstream iss(buf);
		if(first){
			string tmp;
			int cnt = 0;
			for(int i=0; iss>>tmp; i++){
				if(tmp[0] == 'e' || tmp == "ming"){
					marker.push_back(true); cnt++;
				}else{ marker.push_back(false);}
			}
			_Epost.resize(cnt);
			first = false;
		}else{
			double tmp;
			int cnt = 0;
			int iter;
			iss>>iter;
			if(iter > _ctrl->_bi){
				for(int i=1; iss>>tmp; i++){
					if(marker[i]){
						_Epost[cnt].push_back(tmp);
						cnt++;
					}
				}
			}
		}
	}
	ifs.close();
}

void EModel::WriteParam(ostream & ofs, Input & con, bool header){
	int L = _param->l();
	int N = _param->n();
	if(header){
		ofs<<"iter\tllh\tmaxllh\tlntf\teopt\tpave";
		for(int i=0; i< L*4; i++){ ofs<<"\t"<<"e"<<i; }
		for(int i=0; i< L*4; i++){ ofs<<"\t"<<"d"<<i; }
		for(int i=0; i< N; i++) ofs<<"\tge"<<i;
		for(int i=0; i< N; i++) ofs<<"\tg"<<i;
		ofs << "\tgsum";
		for(int i=0; i< N; i++) ofs<<"\tW"<<i;
		ofs<<"\tsigE";
		ofs<<endl;
		
	}else{
		ofs<<con._iter<<"\t"<<_llh<<"\t"<<_maxLlh<<"\t"<<_param->lntf()<<"\t"<<_param->Eopt()<<"\t"<<_param->Pave();
		for(int i=0; i< L; i++){ 
			for(int j=0; j< 4; j++){ ofs<<"\t"<<_param->E(i,j); }}
		for(int i=0; i< L; i++){ 
			for(int j=0; j< 4; j++){ ofs<<"\t"<<_param->D(i,j); }}
		
		for(int i=0; i< N; i++) ofs<<"\t"<<_param->GE(i);
		double gsum =0.0;
		for(int i=0; i< N; i++) { ofs<<"\t"<<_param->G(i); gsum += _param->G(i); }
		ofs<<"\t" << gsum;
		for(int i=0; i< N; i++) ofs<<"\t"<<_param->W(i);
		ofs<<"\t"<<_param->sigE();
		ofs<<endl;
	}
}
void EModel::UpdateModel(int i,int j, double jp){
	int Dij = _param->D(i,j);
	if(Dij == -1){
	}else if(Dij == 0){ _param->Birth(i,j,jp);
	}else if(Dij == 3){ _param->Death(i,j,jp);
	}else{
		if(_rng.uniform(0,1)>0.5){ _param->Birth(i,j,jp);
		}else{ _param->Death(i,j,jp); }
	}
}
bool EModel::UpdateSigE(){
	CheckPoint1();
	_param->SetSigE( _pwSigE->Propose(_param->sigE()) );
	double aft = CalcLlh();
	_trySigE++;
	if(Accept(aft-_llh)){
		_accSigE++; _llh = aft;
		if(_maxLlh < _llh) _maxLlh = _llh;
		return(true);
	}else{
		Rollback1();
	}
	return(false);
}
bool EModel::UpdateLntf(){
	// Update the log( [TF] )
	double bef = _param->CalcLlh_wrtPostboundFreqsOnly(); //CalcLlh(); //_llh
	
	double oldTF = _param->lntf();
	
	//CheckPoint1();
	
	double v = _pwLntf->Propose(_param->lntf());
	
	double amountChanged = v - _param->lntf();
	
	for (int m=0; m<_param->n(); m++) {
		if (!_param->R(m)) {
			_param->SetG(m, _param->G(m) + amountChanged);
		}
	}
	
	_param->SetLntf(v);
	
	
	for(int m=0; m<_param->n(); m++){
		_param->UpdateP(m);
	}
	_param->Normalize();
	
	double aft = _param->CalcLlh_wrtPostboundFreqsOnly(); //CalcLlh();
	_tryLntf++;
	if(Accept(aft-bef)){
		_accLntf++;
		//_llh = aft; 
		if(_llh > _maxLlh) _maxLlh = _llh;
		return(true);
	}else{
		//	_llh = bef;
		//	Rollback1();
		_param->SetLntf( oldTF );
		for(int m=0; m<_param->n(); m++){
			if (!_param->R(m)) {
				_param->SetG(m, _param->G(m) - amountChanged);
			}
			_param->UpdateP(m);
		}
		_param->Normalize();	
	}
	return(false);
}
bool EModel::UpdateEopt(){
	// This appears necessary because G[] is used in likelihood function, while GE[] is not
	double bef = CalcLlh();
	
	CheckPoint1();
	double v = _pwEopt->Propose(_param->Eopt());
	_param->SetEopt(v);
	
	for(int m=0; m<_param->n(); m++){
		_param->UpdateGE(m);
		//	if(!_param->R(m)){ 
		// If not a reference oligo, overwrite G[] with GE[]
		//		_param->SetG(m, _param->GE(m)); 
		//	}
		_param->UpdateP(m);
	}
	_param->Normalize();
	double aft = CalcLlh();
	_tryEopt++;
	
	//	cout << "Proposed value " << v << " with LR " << aft << ", " << bef << endl;
	
	if(Accept(aft-bef)){
		_llh = aft; if(_llh > _maxLlh) _maxLlh = _llh;
		_accEopt++;
		return(true);
	}else{
		Rollback1();
	}
	return(true);
}
bool EModel::UpdateEoptE(double v){
	_param->SetEopt(v);
	double bef = 0;
	double aft = 0;
	double sig = _param->sigE();
	bool isAccepted;
	
	for(int m=0; m<_param->n(); m++){
		double g = _param->G(m);
		bef += DLogNorm(g,_param->GE(m),sig);
		_param->UpdateGE(m);
		aft += DLogNorm(g,_param->GE(m),sig);
	}
	isAccepted = Accept(aft-bef);
	
	_tryEopt++;
	
	if (isAccepted) {
		_accEopt++;
	}
	
	return(isAccepted);
}
bool EModel::UpdateG(int iter){
	// _G is binding energy, _GE is binding energy by additive model
	// DeltaG = RT * ln[ Kd ]
	// DeltaG = ln[ Kd ]
	// Kd = exp( DeltaG )
	static int variancesEstablished = 0;
	static double varianceFactor = _ctrl->_g_win;
	
	CheckPoint2();
	
	double bef2 = _param->CalcLlh(); // _llh;
	
	double bef = _param->CalcLlh_wrtPostboundFreqsOnly(); // _llh;
	double aft = bef;
	double ming = _pwG->_max;
	
	double oldG;
	
	for (int m=0; m<_param->n(); m++) {
	    //int m = int(_rng.uniform(0,_param->n()));
		
		if ( (_param->G(m) < ming) ) {
		    ming = _param->G(m);
		}
		
		if (!_param->R(m)) {
			// If not a reference oligo
			//if ( _rng.uniform(0,1) > 0.2 ) continue;
			
			oldG = _param->G(m);
			
			//CheckPoint1();
			
			if (iter >= 500) {
				if (variancesEstablished == 0) {
					variancesEstablished = 1;
				}
			}
			
			_param->SetG(m, ( oldG + ( _rng.normal( 0.0, varianceFactor ) ) ) );
			_param->UpdateP(m); // gets binding probabilities from binding energies (P=binding prob; G=binding energy)
			_param->Normalize(); // this updates the frequencies after binding (omega^b's), based on binding probabilities (P's = \phi's)
			
			aft = _param->CalcLlh_wrtPostboundFreqsOnly(); //CalcLlh();
			
			_tryG++;
			if (Accept(aft - bef)) {
				bef = aft;
				_accG++;
			} else {
				_param->SetG( m, oldG );
				_param->UpdateP(m);
				//		    Rollback1();
			}
	    }
	}
	
	_param->Normalize();
	CheckPoint1();
	
	// Blocking this...
	//	if( !UpdateEoptE( ming ) ) { Rollback1(); };
	
	aft = CalcLlh();
	if(Accept(aft - bef2)){
		_llh = aft;
		if(_maxLlh < bef2) _maxLlh = bef2;
		return(true);
	}else{ 
		_llh = bef2;
		Rollback2();
	}
	return(false);
}

bool EModel::UpdatePreBoundFreqs(){
	// _G is binding energy, _GE is binding energy by additive model
	
	double bef = _param->CalcLlh_wrtFreqsOnly();
	double aft = bef;
	
	double oldFreq;
	
	for (int m=0; m<_param->n(); m++) {
		oldFreq = _param->W(m);
		
		_param->SetW(m, _pwW->Propose( _param->W(m) ) );
		
		_param->NormalizeW();
		_param->Normalize();
		
		aft = _param->CalcLlh_wrtFreqsOnly();
		
		_tryW++;
		
		if (Accept(aft - bef)) {
		    bef  = aft;
		    _llh = aft;
		    _accW++;
		} else {
			_param->SetW(m, oldFreq);
		}
	}
	
	return(true);
}

bool EModel::UpdateE(){
	// This calls updateE(i,j) multiple times
	CheckPoint2();
	double bef = _param->CalcLlh_wrtEnergyModelOnly();
	double oldEij = 0;
	
	for(int i=0; i<_param->l(); i++){
		for(int j=0; j<4 ;j++){
			oldEij = _param->E(i,j);
			//CheckPoint1();
			
			if(UpdateE(i,j)){
			}else{ 
				_param->SetE(i,j,oldEij); 
				
			} // Rollback1();}
		}
	}
	
	for(int m=0; m<_param->n(); m++){
		_param->UpdateGE(m);
	}
	//	_tryE++;
	
	double aft = _param->CalcLlh_wrtEnergyModelOnly(); // CalcLlh();
	if(Accept(aft-bef)){
		//		_accE++;
		_llh = aft;
		if(_maxLlh < bef) _maxLlh = bef;
		return(true);
	}else{
		Rollback2();
	}
	return(false);
}

bool EModel::UpdateEscale(){
	// Propose scaling all E's by 
	CheckPoint2();
	double bef = _param->CalcLlh_wrtEnergyModelOnly();
	double oldEij = 0;
	
	double change = ( _rng.normal( 1.0, 0.015 ) );
	
	for(int i=0; i<_param->l(); i++){
		for(int j=0; j<4 ;j++){
			oldEij = _param->E(i,j);
			
			if (oldEij != 0.0) { 
				_param->SetE(i,j, (oldEij * change) ); 
			}
		}
	}
	
	for(int m=0; m<_param->n(); m++){
		_param->UpdateGE(m);
	}
	_tryEscale++;
	
	double aft = _param->CalcLlh_wrtEnergyModelOnly(); // CalcLlh();
	if(Accept(aft-bef)){
		_accEscale++;
		_llh = aft;
		if(_maxLlh < bef) _maxLlh = bef;
		return(true);
	}else{
		Rollback2();
	}
	return(false);
}

bool EModel::UpdateE(int i,int j){
	// Core updater for energy model; called by UpdateModelZeroOne()
	
	// if D(i,j) == 0, set E(i,j) = 0 too.  Otherwise propose something.
	
	int dij = _param->D(i,j);
	if(dij < 0) return(false);
	
	// Propose a change to E
	if(dij == 0) {
		_param->SetE(i,j,0);
		return(false);
	} else {
		_param->SetE(i,j, _pwE->Propose(_param->E(i,j)));
	}
	
	double bef = 0;
	double aft = 0;
	double sig = _param->sigE();
	
	// Evaluate the effect of changing energy E(i,j) over all m sequences:
	
	_tryE++;
	for(int m=0; m<_param->n(); m++){
	    if(_param->S(m,i) == j){
			double g = _param->G(m);
			// Evaluate as log normal prob of G() on a lognorm dist with mu=GE and sig=sigE
			bef += DLogNorm(g,_param->GE(m),sig);
			// Now evaluate given the new value of E
			_param->UpdateGE(m);
			aft += DLogNorm(g,_param->GE(m),sig);
		}
	}
	
	bool accept = Accept(aft-bef);
	if (accept) _accE++;
	
	return(accept);
}
bool Parameter::UpdateModelZeroOne(int i, int j, double jp){
	int dij = D(i,j);
	
	if(dij < 0) return(false);
	
	if(dij == 0){
		// If at least one other state has a zero energy term, flip the indicator and update E(i,j)...
		if(CntD(i,0) > 1){
			SetD(i,j,1);
			SetE(i,j,E(i,j) + _rng.uniform(0,jp));
		}else{ return(false); }	
	}else{
		// If the energy parameter is on and is <= jp, set it to 0...
		// [BLOCK] Jason
		//	if(E(i,j) <= jp){
		SetD(i,j,0);
		SetE(i,j,0);
		//	return(false); // J - skip the iterative update to E(i,j) - check
		//	}else{ return(false);}
	}
	for(int m=0; m< n(); m++){
		if(S(m,i)==j) {
			UpdateGE(m);
		}
	}
	//Normalize();
	return(true);
}
bool EModel::Accept(double dllh){
	if(pow(exp(dllh),1/_ctrl->_T) >= _rng.uniform(0,1)){ 
		return(true);
	}
	return(false);
}

void EModel::UpdateModelZeroOne(bool error) {
    int L = _param->l();
	double oldEij = 0;
	double olderEij = 0;
	double oldDij = 0;
	
	double beforeLikelihood = _param->CalcLlh_wrtEnergyModelOnly(); //CalcLlh();
	
    // for every site (L)
    for (int ii=0; ii<L; ii++) {
        int i = int(_rng.uniform(0,L));
		
        for (int j=0; j<4; j++) {
			// CheckPoint2();
			oldDij   = _param->D(i,j);
			olderEij = _param->E(i,j);
			
			// If updatemodelzeroone returns a valid condition for updating,
            if (_param->UpdateModelZeroOne(i,j,_ctrl->_jpa)) {
				
				// Updating E(i,j) _du # of times...
				for (int d=0; d < _ctrl->_du; d++) {
					//  CheckPoint1();
					oldEij = _param->E(i,j);
					
					// Do UpdateE()
                    if (!UpdateE(i,j)) {
                        //Rollback1();
						_param->SetE(i, j, oldEij);
						
						for(int m=0; m<_param->n(); m++){
							if(_param->S(m,i) == j){
								_param->UpdateGE(m);
							}
						}
                    }
                } //else { continue; } ?
				
				double aft = _param->CalcLlh_wrtEnergyModelOnly(); //CalcLlh();
				
                _tryJ++;
				
                if (Accept(aft-beforeLikelihood)) {
                    _accJ++;
                    _llh = aft;
                    if (_maxLlh < _llh) _maxLlh = _llh;
                } else {
					_param->SetD(i,j, oldDij);
					_param->SetE(i,j, olderEij);
					
					for(int m=0; m< _param->n(); m++){
						if(_param->S(m,i)==j) {
							_param->UpdateGE(m);
						}
					}
                    //Rollback2();
                }
            }
        }
    }
}
void EModel::MCMC(string & output, string & method){
	
	ofstream ofs(output.c_str());
	_ctrl->_iter = 0;
	//ostream & ofs = cout;
	_accG=0; _tryG=0;
	_accW=0; _tryW=0; 
	_accE=0; _tryE=0; _tryEscale=0; _accEscale=0;
	_accLntf=0; _tryLntf=0;
	_accEopt=0; _tryEopt=0;
	_accSigE=0; _trySigE=0; 
	_accJ=0; _tryJ=0;
	
	// Initialize energy terms to true values for analysis of simulation data;
	/*	double energies[40] = { 2.803705522, 5.633085248, 0, 4.90309802, 4.747630578, 0, 5.560362682, 5.454659549, 4.666047779, 11.43694818, 0, 10.9705549, 1.907782226, 6.556725193, 3.825926047, 0, 1.911597287, 7.275884102, 0, 6.25191498, 1.973629056, 6.425912124, 0, 5.677561445, 11.17385183, 5.490100134, 0, 8.383954159, 2.501065174, 0, 6.544631657, 4.796237726, 7.952680067, 10.98280819, 0, 6.017446107, 2.199913253, 4.746306106, 0.537819998, 0 };
	 for (int i=0; i<10; i++) {
	 for (int j=0; j<4; j++) {
	 _param->SetE( i, j, energies[(i*4)+j] );
	 }
	 }
	 
	 // Now update GEs...
	 for(int m=0; m< _param->n(); m++){
	 _param->UpdateGE(m);
	 //	cout << "Oligo " << m << " has energy: " << _param->GE(m) << endl;
	 }
	 */
	
	WriteParam(ofs,*_ctrl,true);
	_llh = CalcLlh(); _maxLlh = _llh;
	while(_ctrl->next()){
		
		UpdateLntf(); // Update the log( [TF] )
		if (method.substr(0,2) == "ee") {
			UpdateEscale();
		}
		
		UpdatePreBoundFreqs();
		
		if(method.substr(0,1) == "f"){
			UpdateG(_ctrl->_iter);
			_param->SetSigE(0);
		}else if(method.substr(0,2) == "ee"){
			UpdateG(_ctrl->_iter);
			
			UpdateSigE();
			UpdateE();
			if(!_fixdmat){
				UpdateModelZeroOne(true); // call reversible jump
			}
		}else if(method.substr(0,2) == "e0"){
			// Not implemented.
		}
		
		UpdateLntf();
		if (method.substr(0,2) == "ee") {
			UpdateEscale();
		}
		
		if(_ctrl->_iter >= _ctrl->_bi && _ctrl->_iter % 5 == 0) {
			_llh = CalcLlh();
			WriteParam(ofs,*_ctrl,false);
		}
		if(_ctrl->_iter % 100 == 0){
			// Output % accepted moves for last X moves...
			cerr<<"llh= "<<_llh<<" max= "<<_maxLlh;
			cerr<<" Lntf= "<<_accLntf/(double)(_tryLntf);
			cerr<<" G= "<<_accG/(double)(_tryG);
			cerr<<" W= "<<_accW/(double)(_tryW);
			cerr<<" E= "<<_accE/(double)(_tryE);
			cerr<<" Esc= "<<_accEscale/(double)(_tryEscale);
			cerr<<" Eopt= "<<_accEopt/(double)(_tryEopt);
			cerr<<" SigE= "<<_accSigE/(double)(_trySigE);
			cerr<<" RJ= "<<_accJ/(double)(_tryJ); // << " of " << _tryJ;
			cerr<<endl;
			_accW=0; _tryW=0;
			_accG=0; _tryG=0;
			_accE=0; _tryE=0; _accEscale=0; _tryEscale=0;
			_accLntf=0; _tryLntf=0;
			_accEopt=0; _tryEopt=0;
			_accSigE=0; _trySigE=0;
			_accJ=0; _tryJ=0;
			
		}
	}
	ofs.close();
}

double Parameter::CalcLlh(){
	double llh = LogFactorial(Btot()) + LogFactorial(Ctot());
	
	for(int i =0; i<n(); i++){
		int b = B(i);
		int c = C(i); 
		
		llh += b * logarithm(F(i)) - LogFactorial(b);
		llh += c * logarithm(W(i)) - LogFactorial(c);
		
		double var = _sigE;
		if(var > 0) llh += DLogNorm(G(i),GE(i), var);
	}
	
	return(llh);
}

double Parameter::CalcLlh_wrtEnergyModelOnly(){
	double llh = 0;
	
	for(int i =0; i<n(); i++){
		double var = _sigE ;
		
		if(var > 0) llh += DLogNorm(G(i),GE(i), var );
	}
	
	return(llh);
}

double Parameter::CalcLlh_wrtPostboundFreqsOnly(){
	double llh =0; 
	
	for(int i =0; i<n(); i++){
		int b = B(i);
		double var = _sigE;
		
		llh += b * logarithm(F(i)); 
		if(var > 0) llh += DLogNorm(G(i),GE(i), var ); 
		
	}
	
	return(llh);
}

double Parameter::CalcLlh_wrtFreqsOnly(){
	// Leaves out scaling factors dependent on counts, etc.
	
	double llh = 0;
	
	for(int i =0; i<n(); i++){
		int b = B(i);
		int c = C(i);
		
		llh += b * logarithm(F(i));
		llh += c * logarithm(W(i));
		
		//	if(_sigE > 0) llh += DLogNorm(G(i),GE(i),_sigE);
	}
	
	return(llh);
}

double Parameter::CalcLlh(int k){
	// count after binding x log(freq after binding)
	return( ( _B[k] * logarithm(_F[k]) ) + ( _C[k] * logarithm(_W[k]) ) );
}
double EModel::CalcLlh(){
	return( _param->CalcLlh());
}


