#include "Util.h"

using namespace std;

int logLUTsize;
int logMode;
double logFloor;
double logMin, logMax, logIncr;
double *logLookupTable;

void initLookup(int theN) {
	logLUTsize = theN;
	logFloor = log(1e-70);
	logLookupTable = new double[logLUTsize];
	precomputeLogTable(logLookupTable);
}

void precomputeLogTable(double *LUT) { 
	logMin = 0.00000000000001;
	logMax = 5.000000000;
	logIncr = (logMax-logMin)/logLUTsize;
	
	for(int i=0;i<logLUTsize;i++) { 
		LUT[i] = log((logMin + (i*logIncr))); //save the log value 
	}
	
	logMode = 1;
} 

double logarithm(double val) { 
	if (logMode==0) { return log(val); }
	if (val == 1e-70) { return logFloor;}
	if ((val > logMax) || (val < logMin)) { return log(val); }
	static int x1,x2;
	static double y1,y2, theX1, theX2;
	
	x1 = static_cast <int>( (val-logMin)/logIncr );
	x2 = x1+1;
	theX1 = logMin + (x1*logIncr);
	theX2 = logMin + (x2*logIncr);
	
	y1 = logLookupTable[x1]; y2 = logLookupTable[x2];
	
	return y1 + (val-theX1)*(y2-y1)/(theX2-theX1);
} 

int fexist(const char *filename ) {
	struct stat buffer ;
	if ( stat( filename, &buffer )==0 ) return 1 ;
	return 0 ;
}

vector<int> RevComp(vector<int> x){
	vector<int> res(x.size());
	for(int i=0; i < x.size(); i++){
		res[x.size()-i-1] = RevComp(x[i]);
	}
	return(res);
}

int RevComp(int x){
	if(x==0) return(3);
	if(x==1) return(2);
	if(x==2) return(1);
	if(x==3) return(0);
}

double DLogNorm(double x, double mu, double sigma){
	if(sigma == 0) return(0);
	double tmp = -logarithm(sigma*sqrt(2*PI)) -(x-mu)*(x-mu)/(2*sigma*sigma);	
	if(tmp > 0) tmp = 0;
	return(tmp);	
}
double DNorm(double x, double mu, double sigma){
	
	if(sigma == 0) return(0);
	double tmp = 1/(sigma*sqrt(2*PI))*exp(-(x-mu)*(x-mu)/(2*sigma*sigma));	
	if(tmp > 1) tmp = 1;
	return(tmp);	
}

vector<int> SampleIdx(int N, int n){
	vector<int> res(N); 
	for(int i=0; i<N; i++) res[i] = i;
	Shuffle(res);
	if(n > 0 && n < N){
		vector<int> rr(n);
		for(int i=0; i<n; i++){
			rr[i] = res[i];
		}	
		return(rr);
	}else{
		return(res);
	}
}

vector<int> DiffIdx(int n, vector<int> & v2){
	vector<int> res;
	for(int i=0; i<n; i++){
		bool found = false;
		for(int j=0; j<v2.size(); j++)
			if(i == v2[j]) found = true;	
		if(!found)
			res.push_back(i);
	}
	return(res);	
}
void Shuffle(vector<int> & v){
	int n = v.size();
	for(int i=0; i< n - 1; i++){
		int r=RandIdx(i+1, n -1);
		int tmp = v[i]; v[i] = v[r];v[r]=tmp;
	}
}
double LogCombi(int n, int k){
	return(LogFactorial(n)-LogFactorial(k)-LogFactorial(n-k));
}
vector<double> GenLogFactorial(int n){
	vector<double> res(n);
	for(int i=0; i<= n ;i++){
		res[i] = LogFactorial(i);
	}
	return(res);
}
double LogFactorial(int n) {
	double sum = 0;
	int count = 0;
	if(n <= 2){
		while(count < n){
			count += 1;
			sum += log(count);
		}
	}else{
		// Stirling's Approximation
		sum = n*logarithm(n)-n+logarithm(sqrt(2*PI*n));	
	}
	return sum;
} 
double Mean(vector<double> & in){
	double sum = 0;
	for(int i=0; i< in.size(); i++)
		sum += in[i];
	return(sum/in.size());
}
int RandIdx(int from, int to){
	int range=(to-from+1); 
	return(from+ rand() % range); 
}

double Rand(double from, double to){
	//   	srand((unsigned)time(0)); 
	double range=(to-from); 
	return(from+range*rand()/RAND_MAX); 
}
/*
 template < class T >
 string ToString(const T &arg) {
 ostringstream	out;
 out << arg;
 return(out.str());
 }
 */
string ToString(const int arg) {
	ostringstream	out;
	out << arg;
	return(out.str());
}

string ToString(vector<int> arg, string sep) {
	ostringstream	out;
	for(int i=0; i< arg.size(); i++)
		out << arg[i]<<sep;
	return(out.str());
}
void Tokenize(const string& str, vector<string>& tokens, const string& delimiters)
{
    // Skip delimiters at beginning.
    string::size_type lastPos = str.find_first_not_of(delimiters, 0);
	
	// Find first "non-delimiter".
    string::size_type pos     = str.find_first_of(delimiters, lastPos);
	
    while (string::npos != pos || string::npos != lastPos)
    {
        // Found a token, add it to the vector.
        tokens.push_back(str.substr(lastPos, pos - lastPos));
        // Skip delimiters.  Note the "not_of"
        lastPos = str.find_first_not_of(delimiters, pos);
        // Find next "non-delimiter"
        pos = str.find_first_of(delimiters, lastPos);
    }
}

template <class T>
void PrintVec(vector<T> & v, string deli){
	for(int i=0; i< v.size(); i++)
		cout<<v[i]<<deli;
}
void PrintMat(FILE *fh, double mat[]){
	for(int i=0; i<4; i++){
		for(int j=0; j<4; j++){
			fprintf(fh,"%.4f\t",mat[4*i+j]);
		}
		fprintf(fh,"\n");
	}
}



void QtoP (double Q[], double t, double P[]){
	int n=4;
	for(int i=0; i< n*n; i++) P[i] = 0.0;
	for(int i=0; i< n; i++) P[n*i+i] = 1;
	double Qt[n*n];	
	double Qt2[n*n];	
	double Qt3[n*n];	
	double Qt4[n*n];	
	double Qt5[n*n];	
	double Qt6[n*n];	
	
	Scale(Q,t,Qt);
	Multiply(Qt,Qt,Qt2);
	Multiply(Qt2,Qt,Qt3);
	Multiply(Qt3,Qt,Qt4);
	Multiply(Qt4,Qt,Qt5);
	Multiply(Qt5,Qt,Qt6);
	
	
	for(int i=0; i< n*n; i++){
		P[i] += Qt[i] + Qt2[i]/2 + Qt3[i]/6 + Qt4[i]/24 + Qt5[i]/120 + Qt6[i]/(120*6);
	}
}

void Scale(double A[], double B, double C[]){
	int n = 4;
	for(int i=0; i< n*n ; i++)
		C[i] = A[i] * B;
}

void Multiply(double A[], double B[], double C[]){
	int n = 4;
	for(int i=0; i< n*n; i++) C[i] = 0.0;
	for(int i=0; i< n; i++) 
		for(int j=0; j< n; j++)
			for(int k=0; k< n; k++)
				C[n*i+j] += A[n*i+k] * B[n*k+j];
}

void QtoPi (double Q[], double pi[], int n, double *space){
	/* from rate matrix Q[] to pi, the stationary frequencies:
	 Q' * pi = 0     pi * 1 = 1
	 space[] is of size n*(n+1).
	 */
	int i,j;
	//double *T = space;      /* T[n*(n+1)]  */
	double T[n*(n+1)];
	
	for(i=0;i<n+1;i++) T[i]=1;
	for(i=1;i<n;i++) {
		for(j=0;j<n;j++)
			T[i*(n+1)+j] =  Q[j*n+i];     /* transpose */
		T[i*(n+1)+n] = 0.;
	}
	Matinv(T, n, n+1, pi);
	for(i=0;i<n;i++) pi[i] = T[i*(n+1)+n];
}


void Matinv (double x[], int n, int m, double space[])
{
	/* x[n*m]  ... m>=n
	 space[n].  This puts the fabs(|x|) into space[0].  Check and calculate |x|.
	 Det may have the wrong sign.  Check and fix.
	 */
	register int i,j,k;
	int *irow=(int*) space;
	double ee=1.0e-30, t,t1,xmax, det=1;
	
	for(i=0; i<n; i++) irow[i]=-999999;
	
	for(i=0; i<n; i++)  {
		for (j=i,xmax=0; j<n; j++)
			if (xmax<fabs(x[j*m+i]))
            { xmax = fabs(x[j*m+i]); irow[i]=j; }
		det *= x[irow[i]*m+i];
		if (xmax < ee)   {
			printf("\nDet = %.4e close to zero at %3d!\t\n", xmax,i+1);
		}
		if (irow[i] != i) {
			FOR (j,m) {
				t = x[i*m+j];
				x[i*m+j] = x[irow[i]*m+j];
				x[irow[i]*m+j] = t;
			}
		}
		t = 1./x[i*m+i];
		FOR (j,n) {
			if (j == i) continue;
			t1 = t*x[j*m+i];
			FOR(k,m)  x[j*m+k] -= t1*x[i*m+k];
			x[j*m+i] = -t1;
		}
		FOR(j,m)   x[i*m+j] *= t;
		x[i*m+i] = t;
	}                            /* for(i) */
	for (i=n-1; i>=0; i--) {
		if (irow[i] == i) continue;
		FOR(j,n)  {
			t = x[j*m+i];
			x[j*m+i] = x[j*m + irow[i]];
			x[j*m + irow[i]] = t;
		}
	}
	space[0]=det;
}



